package eu.openaire.urls_controller.util;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimaps;
import com.google.common.collect.SetMultimap;
import eu.openaire.urls_controller.configuration.ImpalaConnector;
import eu.openaire.urls_controller.models.Payload;
import eu.openaire.urls_controller.models.UrlReport;
import org.apache.commons.io.FileDeleteStrategy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Component;

import javax.servlet.http.HttpServletRequest;
import java.io.*;
import java.net.HttpURLConnection;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.sql.Types;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

@Component
public class FileUtils {

    private static final Logger logger = LoggerFactory.getLogger(FileUtils.class);

    @Autowired
    private JdbcTemplate jdbcTemplate;

    @Autowired
    private S3ObjectStore s3ObjectStore;

    @Autowired
    private FileDecompressor fileDecompressor;


    public enum UploadFullTextsResponse {successful, unsuccessful, databaseError}

    public String baseFilesLocation;

    public static final String workingDir = System.getProperty("user.dir") + File.separator;


    public FileUtils (@Value("${services.pdfaggregation.controller.baseFilesLocation}") String baseFilesLocation) {
        if ( !baseFilesLocation.endsWith(File.separator) )
            baseFilesLocation += File.separator;

        if ( !baseFilesLocation.startsWith(File.separator) )
            baseFilesLocation = workingDir + baseFilesLocation;

        this.baseFilesLocation = baseFilesLocation;
    }


    /**
     * In each insertion, a new parquet-file is created, so we end up with millions of files. Parquet is great for fast-select, so have to stick with it and merge those files..
     * This method, creates a clone of the original table in order to have only one parquet file in the end. Drops the original table.
     * Renames the clone to the original's name.
     * Returns the errorMsg, if an error appears, otherwise is returns "null".
     */
    public String mergeParquetFiles(String tableName, String whereClause, String parameter) {
        String errorMsg;
        if ( (tableName == null) || tableName.isEmpty() ) {
            errorMsg = "No tableName was given. Do not know the tableName for which we should merger the underlying files for!";
            logger.error(errorMsg);
            return errorMsg;
        }

        // Make sure the following are empty strings (in case another method call this one in the future with a null-value).
        if ( whereClause == null )
            whereClause = "";

        if ( parameter == null )
            parameter = "";
        else
            parameter = " '" + parameter + "'";  // This will be a "string-check", thus the single-quotes.

        try {
            jdbcTemplate.execute("CREATE TABLE " + ImpalaConnector.databaseName + "." + tableName + "_tmp stored as parquet AS SELECT * FROM " + ImpalaConnector.databaseName + "." + tableName + " " + whereClause + parameter);
            jdbcTemplate.execute("DROP TABLE " + ImpalaConnector.databaseName + "." + tableName + " PURGE");
            jdbcTemplate.execute("ALTER TABLE " + ImpalaConnector.databaseName + "." + tableName + "_tmp RENAME TO " + ImpalaConnector.databaseName + "." + tableName);
            jdbcTemplate.execute("COMPUTE STATS " + ImpalaConnector.databaseName + "." + tableName);
        } catch (Exception e) {
            errorMsg = "Problem when executing the \"clone-drop-rename\" queries!\n";
            logger.error(errorMsg, e);
            return errorMsg;
        }

        return null;    // No errorMsg, everything is fine.
    }


    public static DecimalFormat df = new DecimalFormat("0.00");

    // The following regex might be usefull in a future scenario. It extracts the "plain-filename" and "file-ID" and the "file-extension".
    // Possible full-filenames are: "path1/path2/ID.pdf", "ID2.pdf", "path1/path2/ID(12).pdf", "ID2(25).pdf"
    public static final Pattern FILENAME_ID_EXTENSION = Pattern.compile("(?:([^.()]+)/)?((([^/()]+)[^./]*)(\\.[\\w]{2,10}))$");

    private static final int numOfFullTextsPerBatch = 70;   // The HTTP-headers cannot be too large (It failed with 100 fileNames).

    public static final ExecutorService hashMatchingExecutor = Executors.newFixedThreadPool(6);


    public UploadFullTextsResponse getAndUploadFullTexts(List<UrlReport> urlReports, HttpServletRequest request, long assignmentsBatchCounter, String workerId) {
        // The Controller have to request the files from the Worker, in order to upload them to the S3.
        // We will have to UPDATE the "location" of each of those files in the UrlReports and then insert them all into the database.

        if ( request == null ) {
            logger.error("The \"HttpServletRequest\" is null!");
            return UploadFullTextsResponse.unsuccessful;
        }
        String remoteAddr = request.getHeader("X-FORWARDED-FOR");
        if ( (remoteAddr == null) || "".equals(remoteAddr) )
            remoteAddr = request.getRemoteAddr();

        // Get the file-locations.
        AtomicInteger numFullTextsFound = new AtomicInteger();
        AtomicInteger numFilesFoundFromPreviousAssignmentsBatches = new AtomicInteger();
        int urlReportsSize = urlReports.size();

        SetMultimap<String, Payload> allFileNamesWithPayloads = Multimaps.synchronizedSetMultimap(HashMultimap.create((urlReportsSize / 5), 3));    // Holds multiple values for any key, if a fileName(key) has many IDs(values) associated with it.

        final String getFileLocationForHashQuery = "select `location` from " + ImpalaConnector.databaseName + ".payload where `hash` = ? limit 1";
        final int[] hashArgType = new int[] {Types.VARCHAR};

        List<Callable<Void>> callableTasks = new ArrayList<>(6);

        for ( UrlReport urlReport : urlReports )
        {
            callableTasks.add(() -> {
                Payload payload = urlReport.getPayload();
                if ( payload == null )
                    return null;

                String fileLocation = payload.getLocation();
                if ( fileLocation == null )
                    return null;   // The full-text was not retrieved, go to the next UrlReport.

                // Query the payload-table FOR EACH RECORD to get the fileLocation of A PREVIOUS RECORD WITH THE SAME FILE-HASH.
                // If no result is returned, then this record is not previously found, so go ahead and add it in the list of files to request from the worker.
                // If a file-location IS returned (for this hash), then this file is already uploaded to the S3. Update the record to point to that file-location and do not request that file from the Worker.
                String fileHash = payload.getHash();
                if ( fileHash != null ) {
                    String alreadyFoundFileLocation = null;
                    try {
                        alreadyFoundFileLocation = jdbcTemplate.queryForObject(getFileLocationForHashQuery, new Object[] {fileHash}, hashArgType, String.class);
                    } catch (EmptyResultDataAccessException erdae) {
                        // No fileLocation is found, it's ok. It will be null by default.
                    } catch (Exception e) {
                        logger.error("Error when executing or acquiring data from the the \"getFileLocationForHashQuery\"!\n", e);
                        // TODO - SHOULD WE RETURN A "UploadFullTextsResponse.databaseError" AND force the caller to not even insert the payloads to the database??
                        // TODO - The idea is that since the database will have problems.. there is no point in trying to insert the payloads to Impala (we will handle it like: we tried to insert and got an error).
                        // Unless we do what it is said above, do not continue to the next UrlReport, this query-exception should not disrupt the normal full-text processing.
                    }

                    if ( alreadyFoundFileLocation != null ) {   // If the full-text of this record is already-found and uploaded.
                        payload.setLocation(alreadyFoundFileLocation);  // Set the location to the older identical file, which was uploaded to S3. The other file-data is identical.
                        //logger.debug("The record with ID \"" + payload.getId() + "\" has an \"alreadyRetrieved\" file, with hash \"" + fileHash + "\" and location \"" + alreadyFoundFileLocation + "\"."); // DEBUG!
                        numFilesFoundFromPreviousAssignmentsBatches.incrementAndGet();
                        numFullTextsFound.incrementAndGet();
                        return null;   // Do not request the file from the worker, it's already uploaded. Move on. The "location" will be filled my the "setFullTextForMultiplePayloads()" method, later.
                    }
                }

                // Extract the "fileNameWithExtension" to be added in the HashMultimap.
                Matcher matcher = FILENAME_ID_EXTENSION.matcher(fileLocation);
                if ( ! matcher.matches() ) {
                    logger.error("Failed to match the \"fileLocation\": \"" + fileLocation + "\" of id: \"" + payload.getId() + "\", originalUrl: \"" + payload.getOriginal_url() + "\", using this regex: " + FILENAME_ID_EXTENSION);
                    return null;
                }
                String fileNameWithExtension = matcher.group(2);
                if ( (fileNameWithExtension == null) || fileNameWithExtension.isEmpty() ) {
                    logger.error("Failed to extract the \"fileNameWithExtension\" from \"fileLocation\": \"" + fileLocation + "\", of id: \"" + payload.getId() + "\", originalUrl: \"" + payload.getOriginal_url() + "\", using this regex: " + FILENAME_ID_EXTENSION);
                    return null;
                }

                numFullTextsFound.incrementAndGet();
                allFileNamesWithPayloads.put(fileNameWithExtension, payload);    // The keys and the values are not duplicate.
                // Task with ID-1 might have an "ID-1.pdf" file, while a task with ID-2 can also have an "ID-1.pdf" file, as the pdf-url-2 might be the same with pdf-url-1, thus, the ID-2 file was not downloaded again.
                return null;
            });
        }// end-for

        ImpalaConnector.databaseLock.lock();    // The execution uses the database.
        try {    // Invoke all the tasks and wait for them to finish before moving to the next batch.
            List<Future<Void>> futures = hashMatchingExecutor.invokeAll(callableTasks);
            for ( Future<Void> future : futures ) {
                try {
                    Void result = future.get();   // The result is always "null" as we have a "Void" type.
                } catch (Exception e) {
                    logger.error("", e);
                }
            }
        } catch (InterruptedException ie) {    // In this case, any unfinished tasks are cancelled.
            logger.warn("The current thread was interrupted when waiting for the worker-threads to finish checking for already-found file-hashes: " + ie.getMessage());
            // This is a very rare case. At the moment, we just move on with what we have so far.
        } catch (Exception e) {
            String errorMsg = "Unexpected error when checking for already-found file-hashes in parallel! " + e.getMessage();
            logger.error(errorMsg, e);
            return UploadFullTextsResponse.unsuccessful;
        } finally {
            ImpalaConnector.databaseLock.unlock();  // The remaining work of this function does not use the database.
        }

        logger.info("NumFullTextsFound by assignments_" + assignmentsBatchCounter + "  =  " + numFullTextsFound.get() + " (out of " + urlReportsSize + " | about " + df.format(numFullTextsFound.get() * 100.0 / urlReportsSize) + "%).");
        logger.debug("NumFilesFoundFromPreviousAssignmentsBatches  =  " + numFilesFoundFromPreviousAssignmentsBatches.get());

        ArrayList<String> allFileNames = new ArrayList<>(allFileNamesWithPayloads.keySet());
        int numAllFullTexts = allFileNames.size();
        if ( numAllFullTexts == 0 ) {
            logger.warn("No full-text files were retrieved for assignments_" + assignmentsBatchCounter + " | from worker: " + workerId);
            return UploadFullTextsResponse.successful;    // It was handled, no error.
        }

        // Request the full-texts in batches, compressed in a zstd tar file.
        int numOfBatches = (numAllFullTexts / numOfFullTextsPerBatch);
        int remainingFiles = (numAllFullTexts % numOfFullTextsPerBatch);
        if ( remainingFiles > 0 ) {   // Add an extra batch for the remaining files. This guarantees at least one batch will exist no matter how few (>0) the files are.
            numOfBatches++;
            logger.debug("The assignments_" + assignmentsBatchCounter + " have " + numAllFullTexts + " distinct non-already-uploaded fullTexts (total is: " + numFullTextsFound.get() + "). Going to request them from the Worker \"" + workerId + "\", in " + numOfBatches + " batches (" + numOfFullTextsPerBatch + " files each, except for the final batch, which will have " + remainingFiles + " files).");
        } else
            logger.debug("The assignments_" + assignmentsBatchCounter + " have " + numAllFullTexts + " distinct non-already-uploaded fullTexts (total is: " + numFullTextsFound.get() + "). Going to request them from the Worker \"" + workerId + "\", in " + numOfBatches + " batches (" + numOfFullTextsPerBatch + " files each).");

        // Check if one full text is left out because of the division. Put it int the last batch.
        String baseUrl = "http://" + remoteAddr + ":1881/api/full-texts/getFullTexts/" + assignmentsBatchCounter + "/" + numOfBatches + "/";

        // TODO - The worker should send the port in which it accepts requests, along with the current request.
        // TODO - The least we have to do it to expose the port-assignment somewhere more obvious like inside the "application.yml" file.

        String curAssignmentsBaseLocation = baseFilesLocation + "assignments_" + assignmentsBatchCounter + File.separator;
        // Note: the "curAssignmentsBaseLocation"-directory will be created once the first batch subdirectory is called for creation.

        int failedBatches = 0;
        for ( int batchCounter = 1; batchCounter <= numOfBatches; ++batchCounter ) {
            List<String> fileNamesForCurBatch = getFileNamesForBatch(allFileNames, numAllFullTexts, batchCounter);
            HttpURLConnection conn;
            try {
                conn = getConnection(baseUrl, assignmentsBatchCounter, batchCounter, fileNamesForCurBatch, numOfBatches, workerId);
                if ( conn == null ) {
                    failedBatches ++;
                    continue;   // To the next batch.
                }
            } catch (RuntimeException re) {
                // The "cause" was logged inside "getConnection()".
                failedBatches += (1 + (numOfBatches - batchCounter)); // The "failedBatches" will have the previously failedBatches + this one + the remaining batches which will likely fail too, thus, they will not be tested.
                break;  // The rest of the batches will likely fail as well.
            }

            // Get the extracted files.
            String targetDirectory = curAssignmentsBaseLocation + "batch_" + batchCounter + File.separator;
            Path curBatchPath = null;
            try {
                // Create this batch-directory.
                curBatchPath = Files.createDirectories(Paths.get(targetDirectory));
                // The base-directory will be created along with the first batch directory.

                // Save and decompress the zstd file. Iterate over the PDFs and upload each one of them and get the S3-Url.
                String zstdFileFullPath = targetDirectory + "fullTexts_" + assignmentsBatchCounter + "_" + batchCounter + ".tar.zstd";
                File zstdFile = new File(zstdFileFullPath);

                if ( ! saveArchive(conn, zstdFile) ) {
                    failedBatches ++;
                    continue;   // To the next batch.
                }
                //logger.debug("The zstd file has been saved: " + zstdFileFullPath);    // DEBUG!

                // We do not call "conn.disconnect()", since more request are about to be made to the worker, in the near future.

                fileDecompressor.decompressFiles(zstdFileFullPath, curBatchPath);

                String[] extractedFileNames = new File(targetDirectory).list();
                if ( (extractedFileNames == null) || (extractedFileNames.length <= 2) ) {    // The directory might have only two files, the "tar-file" and the "tar.zstd-file", if the full-texts failed to be decompressed or untarred..
                    logger.error("No full-texts' fleNames where extracted from directory: " + targetDirectory);
                    failedBatches ++;
                    continue;   // To the next batch.
                }
                else if ( (extractedFileNames.length -2) != fileNamesForCurBatch.size() ) {
                    logger.warn("The number of extracted files (" + (extractedFileNames.length -2) + ") was not equal to the number of the current-batch's (" + batchCounter + ") files (" + fileNamesForCurBatch.size() + ").");
                    // We do NOT have to find and cross-reference the missing files with the urlReports, in order to set their locations to <null>,
                    // since, in the end of each assignments-batch, an iteration will be made and for all the non-retrieved and non-uploaded full-texts, the app will set them to null.
                }

                uploadFullTexts(extractedFileNames, targetDirectory, allFileNamesWithPayloads);

            } catch (Exception e) {
                logger.error("Could not extract and upload the full-texts for batch_" + batchCounter + " of assignments_" + assignmentsBatchCounter + "\n" + e.getMessage(), e);   // It shows the response body (after Spring v.2.5.6).
                failedBatches ++;
            } finally {
                if ( curBatchPath != null )
                    deleteDirectory(curBatchPath.toFile());
            }
        }   // End of batches.

        updateUrlReportsToHaveNoFullTextFiles(urlReports, true);    // Make sure all records without an S3-Url have < null > file-data (some batches or uploads might have failed).
        deleteDirectory(new File(curAssignmentsBaseLocation));

        if ( failedBatches == numOfBatches ) {
            logger.error("None of the " + numOfBatches + " batches could be handled for assignments_" + assignmentsBatchCounter + ", for worker: " + workerId);
            return UploadFullTextsResponse.unsuccessful;
        } else
            return UploadFullTextsResponse.successful;
    }


    private HttpURLConnection getConnection(String baseUrl, long assignmentsBatchCounter, int batchNum, List<String> fileNamesForCurBatch, int totalBatches, String workerId) throws RuntimeException
    {
        baseUrl += batchNum + "/";
        String requestUrl = getRequestUrlForBatch(baseUrl, fileNamesForCurBatch);
        //logger.debug("Going to request the batch_" + batchNum + " (out of " + totalBatches + ") with " + fileNamesForCurBatch.size() + " fullTexts, of assignments_" + assignmentsBatchCounter + " from the Worker with ID \"" + workerId + "\" and baseRequestUrl: " + baseUrl + "[fileNames]");
        HttpURLConnection conn = null;
        try {
            conn = (HttpURLConnection) new URL(requestUrl).openConnection();
            conn.setRequestMethod("GET");
            conn.setRequestProperty("User-Agent", "UrlsController");
            conn.connect();
            int statusCode = conn.getResponseCode();
            if ( statusCode == -1 ) {
                logger.warn("Problem when getting the \"status-code\" for url: " + requestUrl);
                throw new RuntimeException();
            } else if ( statusCode != 200 ) {
                logger.warn("HTTP-" + statusCode + ": " + getMessageFromResponseBody(conn, true) + "\n\nProblem when requesting the ZstdFile of batch_" + batchNum + " from the Worker with ID \"" + workerId + "\" and requestUrl: " + requestUrl);
                if ( (statusCode >= 500) && (statusCode <= 599) )
                    throw new RuntimeException();   // Throw an exception to indicate that the Worker has problems and all remaining batches will fail as well.
                return null;
            }
        } catch (RuntimeException re) {
            throw re;
        } catch (Exception e) {
            String exMessage = e.getMessage();
            logger.warn("Problem when requesting the ZstdFile of batch_" + batchNum + " of assignments_" + assignmentsBatchCounter + " from the Worker with ID \"" + workerId + "\" and requestUrl: " + requestUrl + "\n" + exMessage);
            if ( exMessage.contains("Connection refused") ) {
                logger.error("Since we received a \"Connection refused\", all of the remaining batches (" + (totalBatches - batchNum) + ") will not be requested!");
                throw new RuntimeException();
            }
            return null;
        }
        return conn;
    }


    private void uploadFullTexts(String[] fileNames, String targetDirectory, SetMultimap<String, Payload> allFileNamesWithPayloads)
    {
        // Iterate over the files and upload them to S3.
        //int numUploadedFiles = 0;
        for( String fileName : fileNames )
        {
            if ( fileName.contains(".tar") ) // Exclude the tar-files from uploading.
                continue;

            // Check if this stored file is related to one or more Payloads from the Set. Defend against malicious file injection. It does not add more overhead, since we already need the "fileRelatedPayloads".
            Set<Payload> fileRelatedPayloads = allFileNamesWithPayloads.get(fileName);
            if ( fileRelatedPayloads.isEmpty() ) {  // In case the "fileName" is not inside the "allFileNamesWithPayloads" HashMultimap.
                logger.error("The stored file \"" + fileName + "\" is not related to any Payload returned from the Worker!");
                continue;
            }

            // Let's try to upload the file to S3 and update the payloads, either in successful file-uploads (right-away) or not (in the end).
            try {
                // Prepare the filename as: "datasourceid/publicationid::hash.pdf"
                // All related payloads point to this exact same file, BUT, may be related with different urlIDs, which in turn be related with different datasourceIDs.
                // This file could have been found from different urlIds and thus be related to multiple datasourceIds.
                // BUT, since the filename contains a specific urlID, the datasourceId should be the one related to that specific urlID.
                // So, we extract this urlID, search the payload inside the "fileRelatedPayloads" and get the related datasourceID (instead of taking the first or a random datasourceID).

                Matcher matcher = FILENAME_ID_EXTENSION.matcher(fileName);
                if ( !matcher.matches() ) {
                    logger.error("Failed to match the \"" + fileName + "\" with the regex: " + FILENAME_ID_EXTENSION);
                    continue;
                }
                // The "matcher.group(3)" returns the "filenameWithoutExtension", which is currently not used.
                String fileNameID = matcher.group(4);
                if ( (fileNameID == null) || fileNameID.isEmpty() ) {
                    logger.error("Failed to extract the \"fileNameID\" from \"" + fileName + "\".");
                    continue;
                }
                String dotFileExtension = matcher.group(5);
                if ( (dotFileExtension == null) || dotFileExtension.isEmpty() ) {
                    logger.error("Failed to extract the \"dotFileExtension\" from \"" + fileName + "\".");
                    continue;
                }

                // This file is related with some payloads, in a sense that these payloads have urls which lead to the same full-text url.
                // These payloads might have different IDs and sourceUrls. But, in the end, the different sourceUrls give the same full-text.
                // Below, we make sure we pick the "datasource" from the payload, which has the same id as the full-text's name.
                // If there are multiple payloads with the same id, which point to the same file, then we can take whatever datasource we want from those payloads.
                // It is possible that payloads with same IDs, but different sourceUrls pointing to the same full-text, can be related with different datasources
                // (especially for IDs of type: "doiboost_____::XXXXXXXXXXXXXXXXXXXXX").
                // It does not really matter, since the first-ever payload to give this full-text could very well be another one,
                // since the crawling happens in multiple threads which compete with each other for CPU time.

                String datasourceId = null;
                String hash = null;
                boolean isFound = false;
                for ( Payload payload : fileRelatedPayloads ) {
                    if ( fileNameID.equals(payload.getId()) ) {
                        datasourceId = payload.getDatasourceId();
                        hash = payload.getHash();
                        isFound = true;
                        break;
                    }
                }

                if ( !isFound ) {  // This should never normally happen. If it does, then a very bad change will have taken place.
                    logger.error("The \"fileNameID\" (" + fileNameID + ") was not found inside the \"fileRelatedPayloads\" for fileName: " + fileName);
                    continue;
                }

                String s3Url = constructFileNameAndUploadToS3(targetDirectory, fileName, fileNameID, dotFileExtension, datasourceId, hash);
                if ( s3Url == null )
                    continue;

                setFullTextForMultiplePayloads(fileRelatedPayloads, s3Url);
                //numUploadedFiles ++;
            } catch (Exception e) {
                logger.error("Could not upload the file \"" + fileName + "\" to the S3 ObjectStore!", e);
            }
            // Else, the record will have its file-data set to "null", in the end of this method.
        }

        //logger.debug("Finished uploading " + numUploadedFiles + " full-texts (out of " + (fileNames.length -2) + " distinct files) from assignments_" + assignmentsBatchCounter + ", batch_" + batchCounter + " on S3-ObjectStore.");
        // (fileNames.length -2) --> minus the zstd and the tar files.
    }


    public String constructFileNameAndUploadToS3(String fileDir, String fileName, String openAireID, String dotFileExtension, String datasourceId, String hash) throws Exception
    {
        if ( datasourceId == null ) {
            logger.error("The retrieved \"datasourceId\" was \"null\" for file: " + fileName);
            return null;
        }

        if ( hash == null ) {
            logger.error("The retrieved \"hash\" was \"null\" for file: " + fileName);
            return null;
        }

        String fileFullPath = fileDir + File.separator + fileName;   // The fullPath to the local file.

        // Use the "fileNameID" and not the "filenameWithoutExtension", as we want to avoid keeping the possible "parenthesis" with the increasing number (about the duplication of ID-fileName).
        // Now we append the file-hash, so it is guaranteed that the filename will be unique.
        fileName = datasourceId + "/" + openAireID + "::" + hash + dotFileExtension;    // This is the fileName to be used in the objectStore, not of the local file!

        return s3ObjectStore.uploadToS3(fileName, fileFullPath);
    }


    public String getMessageFromResponseBody(HttpURLConnection conn, boolean isError) {
        final StringBuilder msgStrB = new StringBuilder(500);
        try ( BufferedReader br = new BufferedReader(new InputStreamReader((isError ? conn.getErrorStream() : conn.getInputStream()))) ) {  // Try-with-resources
            String inputLine;
            while ( (inputLine = br.readLine()) != null )
            {
                if ( !inputLine.isEmpty() )
                    msgStrB.append(inputLine);
            }
            return (msgStrB.length() != 0) ? msgStrB.toString() : null;	// Make sure we return a "null" on empty string, to better handle the case in the caller-function.
        } catch ( IOException ioe ) {
            logger.error("IOException when retrieving the response-body: " + ioe.getMessage());
            return null;
        } catch ( Exception e ) {   // This includes the case, where the "conn.getErrorStream()" returns < null >.
            logger.error("Could not extract the response-body!", e);
            return null;
        }
    }


    private List<String> getFileNamesForBatch(List<String> allFileNames, int numAllFullTexts, int curBatch) {
        int initialIndex = ((curBatch-1) * numOfFullTextsPerBatch);
        int endingIndex = (curBatch * numOfFullTextsPerBatch);
        if ( endingIndex > numAllFullTexts )    // This might be the case, when the "numAllFullTexts" is too small.
            endingIndex = numAllFullTexts;

        List<String> fileNamesOfCurBatch = new ArrayList<>(numOfFullTextsPerBatch);
        for ( int i = initialIndex; i < endingIndex; ++i ) {
            try {
                fileNamesOfCurBatch.add(allFileNames.get(i));
            } catch (IndexOutOfBoundsException ioobe) {
                logger.error("IOOBE for i=" + i + "\n" + ioobe.getMessage(), ioobe);
            }
        }
        return fileNamesOfCurBatch;
    }


    private String getRequestUrlForBatch(String baseUrl, List<String> fileNamesForCurBatch) {
        final StringBuilder sb = new StringBuilder(numOfFullTextsPerBatch * 50);
        sb.append(baseUrl);
        int numFullTextsCurBatch = fileNamesForCurBatch.size();
        for ( int j=0; j < numFullTextsCurBatch; ++j ){
            sb.append(fileNamesForCurBatch.get(j));
            if ( j < (numFullTextsCurBatch -1) )
                sb.append(",");
        }
        return sb.toString();
    }


    public static final int tenMb = (10 * 1_048_576);

    public boolean saveArchive(HttpURLConnection conn, File zstdFile)
    {
        try ( BufferedInputStream inStream = new BufferedInputStream(conn.getInputStream(), tenMb);
              BufferedOutputStream outStream = new BufferedOutputStream(Files.newOutputStream(zstdFile.toPath()), tenMb) )
        {
            int readBytes;
            while ( (readBytes = inStream.read()) != -1 ) {
                outStream.write(readBytes);
            }
            return true;
        } catch (Exception e) {
            logger.error("Could not save the zstd file \"" + zstdFile.getName() + "\": " + e.getMessage(), e);
            return false;
        }
    }


    /**
     * This method updates the UrlReports to not point to any downloaded fullText files.
     * This is useful when the uploading process of the fullTexts to the S3-ObjectStore fails, and we don't want any "links" to locally stored files, which will be deleted.
     * If the "shouldCheckAndKeepS3UploadedFiles" is set to "true", then the payloads which have their file uploaded to the S3-ObjectStore, are excluded.
     * @param urlReports
     * @param shouldCheckAndKeepS3UploadedFiles
     */
    public void updateUrlReportsToHaveNoFullTextFiles(List<UrlReport> urlReports, boolean shouldCheckAndKeepS3UploadedFiles) {
        for ( UrlReport urlReport : urlReports ) {
            Payload payload = urlReport.getPayload();
            if ( payload == null )
                continue;

            if ( shouldCheckAndKeepS3UploadedFiles ) {
                String fileLocation = payload.getLocation();
                if ( (fileLocation == null) || s3ObjectStore.isLocationInStore(fileLocation) )
                    continue;
            }

            // Mark this full-text as not-retrieved, since it will be deleted from local-storage. The retrieved link to the full-text ("actual_url") will be kept, for now.
            payload.setLocation(null);  // This will cause the payload to not be inserted into the "payload" table in the database. Only the "attempt" record will be inserted.
            payload.setHash(null);
            payload.setMime_type(null);
            payload.setSize(null);
            // The id-url record will be called as a new assignment in the future.
        }
    }


    /**
     * Set the fileLocation for all those Payloads related to the File.
     * @param filePayloads
     * @param s3Url
     */
    public void setFullTextForMultiplePayloads(Set<Payload> filePayloads, String s3Url) {
        for ( Payload payload : filePayloads )
            if ( payload != null )
                payload.setLocation(s3Url); // Update the file-location to the new S3-url. All the other file-data is already set from the Worker.
    }


    public boolean deleteDirectory(File directory)
    {
        try {
            org.apache.commons.io.FileUtils.deleteDirectory(directory);
            return true;    // Will return "true" also in case this directory does not exist. So, no Exception will be thrown for that case.
        } catch (IOException e) {
            logger.error("The following directory could not be deleted: " + directory.getName(), e);
            return false;
        } catch (IllegalArgumentException iae) {
            logger.error("This batch-dir does not exist: " + directory.getName());
            return false;
        }
    }


    public boolean deleteFile(String fileFullPathString)
    {
        try {
            FileDeleteStrategy.FORCE.delete(new File(fileFullPathString));
        } catch (IOException e) {
            logger.error("Error when deleting the file: " + fileFullPathString);
            return false;
        }
        return true;
    }


    Lock fileWriteLock = new ReentrantLock(true);

    public String writeToFile(String fileFullPath, String stringToWrite)
    {
        fileWriteLock.lock();
        try ( BufferedWriter bufferedWriter = new BufferedWriter(Files.newBufferedWriter(Paths.get(fileFullPath)), FileUtils.tenMb) )
        {
            bufferedWriter.write(stringToWrite); // This will overwrite the file. If the new string is smaller, then it does not matter.
        } catch (Exception e) {
            String errorMsg = "Failed to create or acquire the file \"" + fileFullPath + "\"!";
            logger.error(errorMsg, e);
            return errorMsg;
        } finally {
            fileWriteLock.unlock();
        }
        return null;
    }

}